/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "general.h"

#include <openssl/txt_db.h>
#include <openssl/conf.h>
#include <openssl/lhash.h>
#include <openssl/asn1.h>

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/bn.h>
#include <openssl/pem.h>
#include <openssl/bio.h>

#ifdef HAVE_OPENLDAP
#include <ldap.h>
#endif

#include "support.h"
#include "db.h"

IMPLEMENT_LHASH_HASH_FN(index_serial_hash, char **)
IMPLEMENT_LHASH_COMP_FN(index_serial_cmp, char **)
IMPLEMENT_LHASH_HASH_FN(index_name_hash, char **)
IMPLEMENT_LHASH_COMP_FN(index_name_cmp, char **)

/* External imported variables */
extern int debug;
extern int verbose;

/* Functions */
X509_CRL *ocspd_get_crl ( URL *url ) {
	BIO *debug = NULL;
	X509_CRL *crl = NULL;

	if( url->proto == OCSP_CRL_PROTO_LDAP ) {
#ifdef HAVE_OPENLDAP
		crl = ocspd_load_crl_ldap(url);
#endif
	} else if (  url->proto == OCSP_CRL_PROTO_FILE ) {
		crl = ocspd_load_crl_file(url);
	} else {
		return NULL;
	}

	if(verbose && crl) {
		debug=BIO_new_fp(stdout,BIO_NOCLOSE);
		X509_CRL_print( debug, crl );
		BIO_free(debug);
	}

	return crl;

}

#ifdef HAVE_OPENLDAP
X509_CRL *ocspd_load_crl_ldap( URL *url ) {
	LDAP	*ld = NULL;
	int  	i,o,rc;
	char	*attrs[] = { "certificateRevocationList;binary", NULL };
	// char	*attrs[] = { "cACertificate;binary", NULL };
	char    *filter = "objectclass=*";
	struct berval **vals = NULL;
	LDAPMessage *res = NULL;
	BIO	*membio = NULL;

	X509_CRL *crl = NULL;

	if (( ld = ldap_init( url->addr, url->port )) == NULL ) {
		syslog( LOG_ERR, "ldap_init failure (%s:%d)",
		     url->addr, url->port );
		return NULL;
	}

	syslog( LOG_INFO, "LDAP usr %s, LDAP pwd %s",
			url->usr, url->pwd );

	if ( ldap_bind_s( ld, url->usr, url->pwd, LDAP_AUTH_SIMPLE ) 
			!= LDAP_SUCCESS ) {
		syslog( LOG_ERR, "ldap_bind_s failure (%s:%d)",
		     url->addr, url->port );
		return NULL;
	}


	syslog( LOG_INFO, "Successfully binded to LDAP server (%s)",
			url->dn);

	// if (( rc = ldap_search_s( ld, url->dn, LDAP_SCOPE_SUBTREE,

	/* We search for the exact match, so LDAP_SCOPE_BASE is used here */
	if (( rc = ldap_search_s( ld, url->dn, LDAP_SCOPE_BASE,
			filter, attrs, 0, &res )) != LDAP_SUCCESS ) {

		if ( rc != LDAP_NO_SUCH_OBJECT ) 
			syslog( LOG_ERR, "LDAP %s object not found", url->dn );
	} else {
		syslog( LOG_INFO, "Search Successfull");
	}

	i = ldap_count_entries( ld, res );
	syslog( LOG_INFO, "Got %d entries", i );


	if((vals = ldap_get_values_len (ld, res, attrs[0])) != NULL ) {

		syslog( LOG_INFO, "vals is not null" );
		for( i=0; vals[i] != NULL; i++ ) {
			int c;

			/* create a new memory bio where pushing the CRL
			 * data and then load using the ocspd_X509_CRL_bio
			 * function */
			membio = BIO_new(BIO_s_mem());
			BIO_write( membio, vals[i]->bv_val, vals[i]->bv_len );
			BIO_set_flags(membio, BIO_FLAGS_MEM_RDONLY);

			/* Load CRL from the membio */
			crl = ocspd_X509_CRL_bio( membio, FORMAT_ASN1 );

			if( crl ) {
				syslog( LOG_INFO, "CRL loaded successfully");
			} else {
				syslog( LOG_INFO, "can not read CRL data");
			}

			/*
			syslog( LOG_INFO, "printing values (%d)! len %ld", i,
			   			vals[i]->bv_len );
			for ( c = 0; c < vals[i]->bv_len; c++ ) {
				printf( "%c", vals[i]->bv_val[c] );
			}
			*/
		}
	} else {
		syslog( LOG_INFO, "No CRL data found!" );
		return NULL;
	}

	/*
	for ( i = 0; i < maxloop; i++ ) {
		LDAPMessage *res;
		int         rc;

		if (( rc = ldap_search_s( ld, sbase, LDAP_SCOPE_SUBTREE,
				filter, attrs, 0, &res )) != LDAP_SUCCESS ) {

			ldap_perror( ld, "ldap_search" );
			if ( rc != LDAP_NO_SUCH_OBJECT ) break;
			continue;

		}

		ldap_msgfree( res );
	}

	fprintf( stderr, " PID=%ld - Search done.\n", (long) pid );
	*/

	ldap_unbind( ld );
	syslog( LOG_INFO, "Successfully unbinded to LDAP server");

	return crl;
}
#endif

X509_CRL *ocspd_load_crl_file( URL *url ) {
	BIO *in = NULL;
	// STACK_OF(X509_REVOKED) *rev = NULL;
	X509_CRL *crl = NULL;

	/* Load CRL */
	if ((in=BIO_new_file( url->addr, "r")) == NULL) {
		return(NULL);
	}

	crl = ocspd_X509_CRL_bio( in, FORMAT_PEM );

	/*
	if (crlform == FORMAT_ASN1)
		crl=d2i_X509_CRL_bio(in,NULL);
	else if (crlform == FORMAT_PEM)
		crl=PEM_read_bio_X509_CRL(in,NULL,NULL,NULL);
	else	{
		BIO_free( in );
		return(NULL);
		}
	*/
	BIO_free( in );

	/*
	rev = X509_CRL_get_REVOKED(crl);

	for(i = 0; i < sk_X509_REVOKED_num(rev); i++) {
		r = sk_X509_REVOKED_value(rev, i);
		BIO_printf(out,"    Serial Number: ");
		i2a_ASN1_INTEGER(out,r->serialNumber);
		BIO_printf(out,"\n        Revocation Date: ","");
		ASN1_TIME_print(out,r->revocationDate);
		BIO_printf(out,"\n");
		X509V3_extensions_print(out, "CRL entry extensions",
		r->extensions, 0, 8);
	}
	*/

	return crl;
}

X509_CRL *ocspd_X509_CRL_bio ( BIO *in, int form ) {

	X509_CRL *crl = NULL;

	if (form == FORMAT_ASN1)
		/* Format is DER encoded */
		crl=d2i_X509_CRL_bio(in,NULL);
	else if (form == FORMAT_PEM)
		/* Format is PEM encoded */
		crl=PEM_read_bio_X509_CRL(in,NULL,NULL,NULL);
	else	{
		return(NULL);
		}

	return crl;
}

int ocspd_verify_cert ( X509_CRL *crl, ASN1_INTEGER *serial ) {

	int ok;
	X509_REVOKED rtmp;

	/* Look for serial number of certificate in CRL */
	rtmp.serialNumber = serial;
	ok = sk_X509_REVOKED_find(crl->crl->revoked, &rtmp);

	/* Not found: OK */
	if(ok == -1) return 1;

	/* We should have an X509_STORE_CTX *ctx */
	/*
	ok = X509_verify_cert( ctx );

	ok = ctx->check_revocation(ctx);

	cert_store=X509_STORE_new();
	X509_STORE_set_default_paths(cert_store);
	X509_STORE_load_locations(cert_store,certsfile,certsdir);
	X509_STORE_set_verify_cb_func(cert_store,verify_callback);
	ERR_clear_error();
	*/

	/* Returns 0 if it does not find the certificate within the
	 * provided CRL */
	return 0;
}


TXT_DB *ocspd_load_db_file( char *dbfile ) {

	TXT_DB *db;
	BIO *in = NULL;
	
	if( (in = BIO_new_file( dbfile, "r" )) == NULL ) {
                syslog(LOG_ERR,"unable to open '%s'\n",dbfile);
                return(NULL);
	}

        db=TXT_DB_read(in,DB_NUMBER);
        if (db == NULL) return(NULL);

        if (!TXT_DB_create_index(db,DB_serial,NULL,
				LHASH_HASH_FN(index_serial_hash),
                		LHASH_COMP_FN(index_serial_cmp)))
                {
                syslog(LOG_ERR,
			"error creating serial number index:(%ld,%ld,%ld)\n",
			db->error, db->arg1, db->arg2);
                return(NULL);
                }

        if (!TXT_DB_create_index(db,DB_name,index_name_qual,
				LHASH_HASH_FN(index_name_hash),
				LHASH_COMP_FN(index_name_cmp)))
                {
                syslog(LOG_ERR,"error creating name index:(%ld,%ld,%ld)\n",
                        db->error,db->arg1,db->arg2);
                return(NULL);
		}

	return( db );
}

char db_get_status( ASN1_INTEGER *serial, OCSPD_CONFIG *conf ) {

	char 	*row[DB_NUMBER],**rrow;
	char 	ret = DB_TYPE_UNK;

	TXT_DB	*db =NULL;
	BIGNUM 	*bn = NULL;

	db = conf->db;

	bn = ASN1_INTEGER_to_BN(serial,NULL);
	row[DB_serial]=BN_bn2hex(bn);
	BN_free(bn);

	if ((row[DB_name] == NULL) || (row[DB_serial] == NULL))
		{
		syslog(LOG_ERR,"db_get_status: memory allocation failure");
		return (DB_TYPE_UNK);
		}

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	if (rrow == NULL)
		{
		goto err;
		}
	else
		{
		ret = rrow[DB_type][0];
		return (ret);
		}

err:
	syslog( LOG_INFO, "db_get_status: certificate %ld not present in db.",
				ASN1_INTEGER_get(serial));
	return (DB_TYPE_UNK);

}

char **lookup_serial(ASN1_INTEGER *ser, OCSPD_CONFIG *conf) {
	int i;
	BIGNUM *bn = NULL;
	TXT_DB *db;
	char *itmp, *row[DB_NUMBER],**rrow;

	db = conf->db;
	for (i = 0; i < DB_NUMBER; i++)
		row[i] = NULL;

	bn = ASN1_INTEGER_to_BN(ser,NULL);
	itmp = BN_bn2hex(bn);

	row[DB_serial] = itmp;
	BN_free(bn);

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	OPENSSL_free(itmp);

	return rrow;
}

char *db_get_revdate( ASN1_INTEGER *serial, OCSPD_CONFIG *conf ) {

	char 	*row[DB_NUMBER],**rrow;
	char 	*ret = NULL;

	TXT_DB	*db;
	BIGNUM 	*bn = NULL;

	db = conf->db;
	bn = ASN1_INTEGER_to_BN(serial,NULL);
	row[DB_serial]=BN_bn2hex(bn);
	BN_free(bn);

	if ((row[DB_name] == NULL) || (row[DB_serial] == NULL))
		{
		syslog(LOG_ERR,"db_get_revdate: memory allocation failure");
		return (NULL);
		}

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	if (rrow == NULL)
		{
		goto err;
		}
	else
		{
		ret = rrow[DB_rev_date];
		return (ret);
		}

err:
	syslog( LOG_INFO, "db_get_revdate: certificate %ld not present in db.",
				ASN1_INTEGER_get(serial));
	return (NULL);

}

int db_set_status(X509 *x509, OCSPD_CONFIG *conf, char MODE)
{
	ASN1_UTCTIME *tm=NULL, *revtm=NULL;
	char *row[DB_NUMBER],**rrow,**irow;
	BIGNUM *bn = NULL;
	TXT_DB *db;
	int ok=-1,i;

	db = conf->db;

	for (i=0; i<DB_NUMBER; i++)
		row[i]=NULL;
	row[DB_name]=X509_NAME_oneline(X509_get_subject_name(x509),NULL,0);
	bn = ASN1_INTEGER_to_BN(X509_get_serialNumber(x509),NULL);
	row[DB_serial]=BN_bn2hex(bn);
	BN_free(bn);

	if ((row[DB_name] == NULL) || (row[DB_serial] == NULL))
		{
		syslog(LOG_ERR,"db_set_status: memory allocation failure\n");
		return(0);
		}

	/* We have to lookup by serial number because name lookup
	 * skips revoked certs
 	 */

	rrow=TXT_DB_get_by_index(db,DB_serial,row);
	if (rrow == NULL)
		{
		if( verbose )
			syslog(LOG_ERR,
				"db_set_status: adding Entry to DB for %s\n",
				row[DB_name]);

		/* We now just add it to the database */
		row[DB_type]=(char *)OPENSSL_malloc(2);

		tm=X509_get_notAfter(x509);
		row[DB_exp_date]=(char *)OPENSSL_malloc(tm->length+1);
		memcpy(row[DB_exp_date],tm->data,tm->length);
		row[DB_exp_date][tm->length]='\0';

		row[DB_rev_date]=NULL;

		/* row[DB_serial] done already */
		row[DB_file]=(char *)OPENSSL_malloc(8);

		/* row[DB_name] done already */

		if ((row[DB_type] == NULL) || (row[DB_exp_date] == NULL) ||
			(row[DB_file] == NULL))
			{
			syslog(LOG_ERR,
				"db_set_status: memory allocation failure");
			return(0);
			}
		strcpy(row[DB_file],"unknown");
		row[DB_type][0]='V';
		row[DB_type][1]='\0';

		if ((irow=(char **)OPENSSL_malloc(sizeof(char *)*(DB_NUMBER+1))) == NULL)
			{
			syslog(LOG_ERR,
				"db_set_status: memory allocation failure");
			goto err;
			}

		for (i=0; i<DB_NUMBER; i++)
			{
			irow[i]=row[i];
			row[i]=NULL;
			}
		irow[DB_NUMBER]=NULL;

		if (!TXT_DB_insert(db,irow))
			{
			syslog(LOG_ERR,
				"db_set_status: failed to update database");
			syslog(LOG_ERR,"db_set_status: TXT_DB error number %ld",
					db->error);
			goto err;
			}

		if( verbose )
			syslog(LOG_INFO,"db_set_status: added Entry to DB %s",
				       irow[DB_serial]);

		ok = db_set_status(x509,conf,MODE);
		return(ok);

		}
	else if (index_name_cmp(row,rrow))
		{
		syslog(LOG_ERR,"db_set_status: name does not match %s",
			   row[DB_name]);
		goto err;
		}
	else if (rrow[DB_type][0]== DB_TYPE_REV && MODE == DB_TYPE_SUS)
		{
		syslog(LOG_ERR,
			"db_set_status: already revoked, serial number %s",
			   row[DB_serial]);
		goto err;
		}
	else if (rrow[DB_type][0]== DB_TYPE_EXP && MODE == DB_TYPE_SUS)
		{
		syslog(LOG_ERR,
			"db_set_status: already expired, serial number %s",
			   row[DB_serial]);
		goto err;
		}
	else
		{
		syslog(LOG_INFO,
			"db_set_status: updating Certificate %s status.",
							rrow[DB_serial]);
		revtm = ASN1_UTCTIME_new();
		revtm=X509_gmtime_adj(revtm,0);

		rrow[DB_type][0]= MODE;
		rrow[DB_type][1]='\0';

		if( MODE == DB_TYPE_SUS ) {
			rrow[DB_rev_date]=(char *)
					OPENSSL_malloc(revtm->length+1);
			memcpy(rrow[DB_rev_date],revtm->data,revtm->length);
			rrow[DB_rev_date][revtm->length]='\0';
		} else {
			rrow[DB_rev_date] = '\0';
		}

		ASN1_UTCTIME_free(revtm);
		}
	ok=1;
err:
	for (i=0; i<DB_NUMBER; i++)
		{
		if (row[i] != NULL) 
			OPENSSL_free(row[i]);
		}
	return(ok);
}

