/* src/net/server.c 
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include "general.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <resolv.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/conf.h>
#include <openssl/txt_db.h>

#include "support.h"
#include "ocsp_response.h"
#include "http_request.h"
#include "server.h"

/* External General Variables */
extern int debug;
extern int verbose;

extern OCSPD_CONFIG *ocspd_conf;

static int *child_list = NULL;
static int max_child = 10;
static int live_childrens = 0;

/* If this is true =1 then we should stop spawning processes */
static int stop_server = 0;

int start_server( char *bind_s, char *port_s, int max_ch_num,
		int queue, long max_size, OCSPD_CONFIG *ocspd_conf ) {

	BIO *bio_socket, *buf_bio;

	HTTP_REQUEST *http_parsed_req = NULL;
	OCSP_REQUEST *OCSP_req = NULL;
	OCSP_RESPONSE *OCSP_resp = NULL;

	char *buf = NULL;
	char tmp_socket[1024];
	char *addr = NULL;
	int addr_len;

	int i = 0;
	long buf_len = 0;
	int lck = -1;

#ifdef _USE_SEMAPHORES
	union semun options;
	struct sembuf lock_it;
	key_t my_key;
#endif

	/* Let's copy the passed pointer of child_lst into the static
	 * variable child_list */
	max_child = max_ch_num;

	/* Let's add for later SSL support */
	SSL_load_error_strings();
	OpenSSL_add_ssl_algorithms();

	/* No max size under 1024 */
	if( max_size < 1024 ) max_size = 1024;
	
	/* Alloc addr for binding to specific address:port */
	addr_len = strlen(bind_s) + strlen(port_s) + 2 ;
	if((addr = (char *) OPENSSL_malloc (addr_len)) == NULL ) {
		syslog(LOG_ERR, "Cannot alloc addr");
		exit( -5 );
	}

	/* We build the "host:port" string to be used */
	sprintf( addr, "%s:%s", bind_s, port_s );

	if( (bio_socket = init_connection( addr )) == NULL ) {
		syslog(LOG_ERR,"Can not setup socket, exit.");
		exit( -6 );
	}
	
#ifdef _USE_FLOCK
	/* We need a file descriptor to make locks on it */
	if( (lck = open("/tmp/ocsp.lck", O_CREAT )) == 0 ) {
		syslog(LOG_ERR, "Impossible to open flock file");
		return 0;
	}
#endif
#ifdef _USE_SEMAPHORES
	/* Let's create the semaphore and check it */
	my_key = ftok(".", 3560);
	lck = semget( my_key, 1, IPC_CREAT | IPC_EXCL | 0600 );
	options.val = 1;
	semctl( lck, 0, SETVAL, options);

	if( semctl(lck,0,GETVAL,0) == 0 ) {
		syslog( LOG_ERR, "Cannot Lock the Semaphore (%d)",lck);
		exit(-1);
	}
#endif
	/* Set the Lock Management into config variable */
	ocspd_conf->lck = lck;

	/* Now let's spaw childrens so that can compete on the accept() */
	spawn_children( &bio_socket, max_child, ocspd_conf, lck);

	// Register signal handlers
	signal( SIGCHLD, handle_sigchld );
	signal( SIGKILL, handle_sigkill );
	signal( SIGTERM, handle_sigkill );
	signal( SIGSTOP, handle_sigkill );

	signal( SIGHUP, handle_sighup );

	while ( pause() ) {
		/* Here we should monitor the child list for dead
		   children */

		if( stop_server > 0 )
			break;

		if( live_childrens < max_child ) {
			wait(0);

			syslog( LOG_ERR, "Some childrens are dead.");

			/* Let's spawn other childrens */
			if( stop_server == 0 )
				spawn_children( &bio_socket, max_child, 
							ocspd_conf, lck);
		}
	}

	// free_lock( ocspd_conf->lck );
	// ocspd_conf->lck = -1;

	return(1);
}

BIO *init_connection ( char *addr ) {

	BIO *bio_socket, *buf_bio;

	/* Set buffered bio */
	buf_bio = BIO_new(BIO_f_buffer());
	if (!buf_bio) {
		syslog( LOG_ERR, "Can not set bufbio %d", __LINE__ );
		return (NULL);
	}

	/* Bind to socket */
	if((bio_socket = BIO_new_accept( addr )) == NULL ) {
		syslog(LOG_ERR, "Cannot bind to %s", addr);
		return (NULL);
	}

	if (verbose) syslog(LOG_INFO, "successfully binded to %s", addr);

        BIO_set_accept_bios(bio_socket, buf_bio);
	buf_bio = NULL;
	if (BIO_do_accept(bio_socket) <= 0)
	{
		syslog( LOG_ERR, "Error setting up accept BIO");
		BIO_free_all( bio_socket );
		BIO_free( buf_bio );

		return( NULL );
	}

	return bio_socket;
}

int handle_connection( BIO **bio_socket, int lck, OCSPD_CONFIG *ocspd_conf ) {

	int pid, i, ret = 0;
	int status = 0;
        BIO *curr_bio = NULL;
	BIO *pbio_socket = NULL;

#ifdef _USE_SEMAPHORES
	struct sembuf lock_it;
#endif

	// struct tms t1,t2;
	// clock_t c1,c2;

	OCSP_REQUEST *ocsp_req = NULL;
	OCSP_RESPONSE *ocsp_resp = NULL;

	pbio_socket = *bio_socket;

	while ( 1 ) {
#ifdef _USE_FLOCK
		if( flock( lck, LOCK_EX ) == 0 ) {
			syslog( LOG_ERR, "Cannot get lock on resource");
			break;
		}
#endif
#ifdef _USE_SEMAPHORES
		/* Set the semaphore */
		lock_it.sem_num = 0;
		lock_it.sem_op = -1;
		lock_it.sem_flg = 0;

		if((semop(lck, &lock_it, 1)) == -1 ) {
			syslog( LOG_ERR, "Can not lock the semaphore (%d)",
				ocspd_conf->lck);
			return(0);
		}
#endif

		// c1 = clock();
		if (BIO_do_accept( pbio_socket ) <= 0) {
			syslog(LOG_INFO, "Error in getting lock (%d)!",
				getpid());
			return 0;
		}
		// c2=clock();

#ifdef _USE_SEMAPHORES
		/* Un-Set the semaphore */
		lock_it.sem_num = 0;
		lock_it.sem_op = 1;
		lock_it.sem_flg = 0;

		if((semop(lck, &lock_it, 1)) == -1 ) {
			syslog( LOG_ERR, "Can not unlock the semaphore");
			return(0);
		}
#endif

		// syslog(LOG_INFO,"Accept User Took: %ld", c2-c1);
		// syslog(LOG_INFO,"Accept System Took: %ld", t2.tms_stime - t1.tms_stime);
		if( verbose )
			syslog(LOG_INFO, "Got lock on file descriptor (%d).",
				getpid());
		curr_bio = BIO_pop( pbio_socket );

		if((ocsp_req = get_ocsp_request( curr_bio )) == NULL ) {
			syslog(LOG_ERR, "Can not parse ocsp request");
			goto err;
		}

		if(( make_ocsp_response(&ocsp_resp, 
					ocsp_req, ocspd_conf)) == 0 ) {
			syslog( LOG_ERR, "Error in generating response" );
			goto err;
		}

		send_ocsp_response( curr_bio, ocsp_resp );

		goto end;

err:
		ocsp_resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		send_ocsp_response( curr_bio, ocsp_resp );

end:
		BIO_free_all(curr_bio);

		if( ocsp_req ) OCSP_REQUEST_free (ocsp_req );
        	if( ocsp_resp) OCSP_RESPONSE_free(ocsp_resp);
	}

	return ret;
}


OCSP_REQUEST *get_ocsp_request( BIO *bio ) {

	int r, have_post, req_len;
	char buf[1024];
	char *header = NULL;

	OCSP_REQUEST *ocsp_req = NULL;

	req_len = 0;
	have_post = 0;
	for( ;; ) {
		r = BIO_gets(bio, buf, 1024);
		if( r <= 0 )
			return (NULL);

		if(!have_post) {
			if( strncmp( buf, "POST", 4) ) {
				syslog(LOG_ERR,"Method is not POST, rejecting");
				return (NULL);
			}
			have_post = 1;
		}

		/* Here we have seen all the header lines */
		if((buf[0] == '\n') || (buf[0] == '\r')) {
			break;
		}
	}

	ocsp_req = d2i_OCSP_REQUEST_bio(bio, NULL);
	if (!ocsp_req) {
		syslog( LOG_ERR, "Error parsing OCSP request\n");
		return(NULL);
	}

	return ocsp_req;
}

int send_ocsp_response(BIO *bio, OCSP_RESPONSE *resp) {

	char http_resp[] =
		"HTTP/1.0 200 OK\r\nContent-type: application/ocsp-response\r\n"
 		"Content-Length: %d\r\n\r\n";

	if (!bio) {
		syslog(LOG_ERR, "Error sending response, bio is NULL");
		return 0;
	}

	BIO_printf(bio, http_resp, i2d_OCSP_RESPONSE(resp, NULL));
	i2d_OCSP_RESPONSE_bio(bio, resp);
	BIO_printf(bio,"\n\n");

	BIO_flush(bio);
	return 1;
}

void handle_sigchld( int sig ) {
	int ret;

	live_childrens--;
	syslog( LOG_ERR, "Live childrens now: %d", live_childrens);

	return;
}

void handle_sigkill( int sig ) {
	int i;

	syslog(LOG_INFO, "Shutting down, master (received signal %d).", sig );

	/* We should stop the server */
	stop_server = 1;

	// Register signal handlers
	signal( SIGCHLD, SIG_IGN );
	signal( SIGTERM, SIG_IGN );

	if(verbose)
		syslog( LOG_INFO, "Sending SIGTERM signal to childrens.");

	kill( 0, SIGTERM );

	while ( 1 ) {
		sleep(1);

		/* We now see if there are waiting childrens */
		i = waitpid( -1, NULL, WNOHANG );

 		if( i < 1 ) {
			/* No more childrens */
			if (verbose)
				syslog(LOG_INFO,
					"All childrens died.");
			break;
		} else {
			/* Let's free all the waiting childrens */
			while( waitpid( -1, NULL, WNOHANG) > 0 );
		}
	}

	/* Let's free the resources */
	free_lock( ocspd_conf->lck );
	ocspd_conf->lck = -1;

	if(verbose)
		syslog(LOG_INFO, "All done. Happy to serve you, bye!");

	return;
}

void handle_sighup( int sig ) {

	/* load crl */
	if((ocspd_conf->crl_data.crl = 
			ocspd_get_crl(ocspd_conf->crl_data.url)) == 0 ) {
		syslog ( LOG_ERR, "can not reload CRL, using the old one.");
		return;
	}

	/* load/init certificates db */
	/*
	 if(( ocspd_conf->db = 
			ocspd_load_db_file(ocspd_conf->db_file )) == NULL ) {

		syslog( LOG_ERR,"can not load/init db (%s)", 
				ocspd_conf->db_file );
		return;
	}
	*/

	syslog( LOG_INFO, "certificate db successfully reloaded (%s)",
			ocspd_conf->crl_data.url->addr );
	return;

}


void free_lock( int lck ) {

	/* Let's free the lock. It is not very important for
	   flock() support but it is needed for semaphores */

	if( lck > 0 ) {
#ifdef _USE_FLOCK
		close(lck);
#endif
#ifdef _USE_SEMAPHORES
		semctl(lck, 0, IPC_RMID, 0);
#endif
		lck = -1;
	}
}

int spawn_children(BIO **bio, int max_child, 
			OCSPD_CONFIG *ocspd_conf, int lck) {

	int i;
	int pid = 0;
	int ppid = 0;

	/* Allocate array for the child list */
	/*
	if ((child_list = (int *) OPENSSL_malloc ( sizeof(int) * max_child ))
			== NULL ) {
		syslog(LOG_ERR, "Memory allocation error: %d", __LINE__ );
		exit(1);
	}
	*/

	syslog( LOG_INFO, "Pre-Spawning till %d processes (live %d)", 
						max_child, live_childrens );

	/* Let's init the client list, -1 means free space */
	for( i = live_childrens ; i < max_child; i++ ) {
		pid = fork();
		if ( pid == 0 ) {
			/* child */
			ppid = getppid();

			// Register signal handlers
			signal( SIGCHLD, SIG_DFL );
			signal( SIGKILL, SIG_DFL );
			signal( SIGTERM, SIG_DFL );
			signal( SIGSTOP, SIG_DFL );
			signal( SIGHUP,  SIG_DFL );

			handle_connection( bio, lck, ocspd_conf );
			raise(SIGTERM);
		} else {
			/* Parent Process */
			/* If an error is present the child is
			   not spawned and the pid value is -1 */
			// child_list[i] = pid;
			live_childrens++;
		}
	}

	if( verbose )
		syslog( LOG_INFO, "Active Childrens %d", live_childrens );

	return 1;
}
